︠8c8e47fb-825c-49c8-99dd-bcdeb19e9aebs︠
# Adapted from http://www.wstein.org/edu/2010/414/projects/hoon_kwon.pdf
N.<A_w,A_x,A_y,A_z,B_w,B_x,B_y,B_z> = QQ[]

# Yields variables i,j,k that comprise the Hamiltonian quaternion basis.
H.<i,j,k> = QuaternionAlgebra(Frac(N),-1,-1)

quatA = A_w + A_x*i + A_y*j + A_z*k
quatB = B_w + B_x*i + B_y*j + B_z*k

# This is a neat trick, but I can't seem to do much else with these QuaternionAlgebra[Element] types.
quatA * quatB

︡64e15cba-c6c1-45f5-b2b2-df3444ceb2bd︡{"stdout":"A_w*B_w - A_x*B_x - A_y*B_y - A_z*B_z + (A_x*B_w + A_w*B_x - A_z*B_y + A_y*B_z)*i + (A_y*B_w + A_z*B_x + A_w*B_y - A_x*B_z)*j + (A_z*B_w - A_y*B_x + A_x*B_y + A_w*B_z)*k\n"}︡{"done":true}︡
︠23f7eb44-3421-4c0a-929f-be948d4623e1s︠

# Constructs a unit quaternion using Euler parameters.
def QuatFromAxisAngle(vec, thetaRad):
    (axisX, axisY, axisZ) = (vec[0], vec[1], vec[2])
    halfTheta = thetaRad / 2
    w = cos(halfTheta)
    x = axisX * sin(halfTheta)
    y = axisY * sin(halfTheta)
    z = axisZ * sin(halfTheta)
    return vector([w, x, y, z])

def AxisAngleFromQuat(quat):
    theta = 2 * arccos(quat[0])
    axis = vector([quat[1], quat[2], quat[3]])
    axis /= sqrt(1 - (quat[0] * quat[0]))
    return vector([theta, axis[0], axis[1], axis[2]])

show( AxisAngleFromQuat(vector([0,0,1,0])) )

︡0ecbd25d-21c2-4861-8cf5-7086aeec5dfd︡{"html":"<div align='center'>$\\displaystyle \\left(\\pi,\\,0,\\,1,\\,0\\right)$</div>"}︡{"done":true}︡
︠4daff251-a764-43db-a414-726e3c5c7c4ds︠


xAxis = vector([1, 0, 0])
yAxis = vector([0, 1, 0])
zAxis = vector([0, 0, 1])

qx90 = QuatFromAxisAngle(xAxis, pi / 2)
qy90 = QuatFromAxisAngle(yAxis, pi / 2)
qz90 = QuatFromAxisAngle(zAxis, pi / 2)
halfSqrt2 = sqrt(2) / 2
assert qx90 == vector([halfSqrt2, halfSqrt2, 0, 0])
assert qy90 == vector([halfSqrt2, 0, halfSqrt2, 0])
assert qz90 == vector([halfSqrt2, 0, 0, halfSqrt2])
show('$x(90^{\circ})$', column_matrix(qx90), '$y(90^{\circ})$', column_matrix(qy90), '$z(90^{\circ})$', column_matrix(qz90))

︡abc86666-3533-4ec3-a35b-48888534103d︡{"html":"<div align='center'>$x(90^{\\circ})$ $\\displaystyle \\left(\\begin{array}{r}\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0 \\\\\n0\n\\end{array}\\right)$ $y(90^{\\circ})$ $\\displaystyle \\left(\\begin{array}{r}\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0 \\\\\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0\n\\end{array}\\right)$ $z(90^{\\circ})$ $\\displaystyle \\left(\\begin{array}{r}\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0 \\\\\n0 \\\\\n\\frac{1}{2} \\, \\sqrt{2}\n\\end{array}\\right)$</div>"}︡{"done":true}︡
︠22884fea-6be7-48f5-91f8-40c3c260d520s︠

# The Hamilton product of two quaternions.
def quat_product(lhs, rhs):
    (r1, v1) = (lhs[0], lhs[1:4])
    (r2, v2) = (rhs[0], rhs[1:4])

    scalarProduct = r1 * r2
    dotProduct = v1.dot_product(v2)
    crossProduct = v1.cross_product(v2)

    resultScalar = scalarProduct - dotProduct
    resultVector = (r1 * v2) + (r2 * v1) + crossProduct

    return vector([resultScalar, resultVector[0], resultVector[1], resultVector[2]])


# Yields the rotation corresponding to the combination of the entire input list of rotation quaternions.
def compose_quats(quat_list):
    return reduce(quat_product, quat_list, vector([1, 0, 0, 0]))


qidentity = vector([1, 0, 0, 0])

assert quat_product(qidentity, qidentity) == qidentity

assert compose_quats([]) == qidentity
assert compose_quats([qidentity]) == qidentity
assert compose_quats([qidentity, qidentity]) == qidentity
︡1e97e0cd-5ed5-43bc-9000-962a8fc0e0dc︡{"done":true}︡
︠9aca8bed-2f91-49e9-ba06-256ccfdcbbe2s︠

(theta_x, theta_y, theta_z) = var('theta_x, theta_y, theta_z')

qx = QuatFromAxisAngle(xAxis, theta_x)
qy = QuatFromAxisAngle(yAxis, theta_y)
qz = QuatFromAxisAngle(zAxis, theta_z)

show(column_matrix(qx), column_matrix(qy), column_matrix(qz))
show(column_matrix(qx.subs(theta_x == pi / 2)), column_matrix(qy.subs(theta_y == pi / 2)), column_matrix(qz.subs(theta_z == pi / 2)))

︡5ee78f23-9360-4307-b8d9-c57cea6d42d4︡{"html":"<div align='center'>$\\displaystyle \\left(\\begin{array}{r}\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\\\\n\\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\\\\n0 \\\\\n0\n\\end{array}\\right)$ $\\displaystyle \\left(\\begin{array}{r}\n\\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\\\\n0 \\\\\n\\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\\\\n0\n\\end{array}\\right)$ $\\displaystyle \\left(\\begin{array}{r}\n\\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n0 \\\\\n0 \\\\\n\\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right)\n\\end{array}\\right)$</div>"}︡{"html":"<div align='center'>$\\displaystyle \\left(\\begin{array}{r}\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0 \\\\\n0\n\\end{array}\\right)$ $\\displaystyle \\left(\\begin{array}{r}\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0 \\\\\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0\n\\end{array}\\right)$ $\\displaystyle \\left(\\begin{array}{r}\n\\frac{1}{2} \\, \\sqrt{2} \\\\\n0 \\\\\n0 \\\\\n\\frac{1}{2} \\, \\sqrt{2}\n\\end{array}\\right)$</div>"}︡{"done":true}︡
︠5bf4d6e2-47a1-4383-b077-c45b86e7d643s︠

qxyz = compose_quats([qx, qy, qz])
show(column_matrix(qxyz))

︡29f8f277-8bb6-4858-b3ad-28a76cabec8d︡{"html":"<div align='center'>$\\displaystyle \\left(\\begin{array}{r}\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) - \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) + \\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) - \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) + \\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right)\n\\end{array}\\right)$</div>"}︡{"done":true}︡
︠c5bd35f9-48e2-4200-a5c3-0a1f61f5ea74s︠

qzxy = compose_quats([qz, qx, qy])
show(column_matrix(qzxy))

︡1f885741-5458-42b7-9838-4d87e86cf8ed︡{"html":"<div align='center'>$\\displaystyle \\left(\\begin{array}{r}\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) - \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) - \\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) + \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) + \\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right)\n\\end{array}\\right)$</div>"}︡{"done":true}︡
︠b9f84b87-52d5-4060-820d-560f519e36fas︠

qzyx = compose_quats([qz, qy, qx])
show(column_matrix(qzyx))

︡3c70d941-98a5-48e3-9f60-4b0c744ea7ee︡{"html":"<div align='center'>$\\displaystyle \\left(\\begin{array}{r}\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) + \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) - \\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n\\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) + \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\\\\n-\\cos\\left(\\frac{1}{2} \\, \\theta_{z}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{y}\\right) + \\cos\\left(\\frac{1}{2} \\, \\theta_{x}\\right) \\cos\\left(\\frac{1}{2} \\, \\theta_{y}\\right) \\sin\\left(\\frac{1}{2} \\, \\theta_{z}\\right)\n\\end{array}\\right)$</div>"}︡{"done":true}︡
︠fa2a196f-26b7-4ba9-a30f-51ca09c0db16s︠


# This is a whole bunch of me trying to figure out how to reorder the terms so the theta variables always appear in the order x,y,z
# See also: http://ask.sagemath.org/question/10641/can-sage-stop-ordering-terms/?answer=15600#post-id-15600
assert qxyz[0].operator() == sage.symbolic.operators.add_vararg
assert qxyz[0].operands()[0].operator() == sage.symbolic.operators.mul_vararg
assert qxyz[0].operands()[0].operands()[0].operator() == cos
assert qxyz[0].operands()[0].operands()[0].operands()[0].operator() == sage.symbolic.operators.mul_vararg
#qxyz[0].operands()[0].operands()[0].operands()[0].operands()[0] == theta_x
#qxyz[0].operands()[0].operands()[0].operands()[0].operands()[1] == 1/2
#qxyz[0].operands()[0].operands()[0].variables()[0] == theta_x
︡5934765e-3145-45b8-9127-6f273f823d28︡{"done":true}︡
︠a407465d-8db9-4d94-8673-84d21e22fd66︠









